from pathlib import Path

import spike2py_reflex as s2pr


def test_5hkz_bi_tss_tss():
    trial_ = '5khz_bi_tss'
    subject_ = 'S4'
    from_command_line = True
    plot=True
    sections = ["mmax_pre", "mmax_post", "tss"]
    study_path = Path('/home/martin/Dropbox/Martin/sketchbook/python/projects/spike2py_reflex/test/data/study5')


def test_process_study():
    study_path = Path('/home/martin/Dropbox/Martin/sketchbook/python/projects/spike2py_reflex/test/data/study5')
    s2pr.process.study(study_path, plot=True)



import matplotlib.pyplot as plt
import matplotlib.patches as patches
import numpy

import spike2py_reflex as s2pr

PLOT_WIDTH_SCALING = 4
PLOT_HEIGHT_SCALING = 2
REFLEX_COLORS = ['m', 'g', 'k']


def plot_single(section):
    """Plot section single reflexes.

    Generates a plot with n columns, where n = 2 * number_muscles, and m rows,
    where m = number of stimulation intensities. For each muscle, there are
    two columns. The first column plots the data extracted for each reflex,
    and also the avg reflex. The second column plots the same data, but with
    the `plot` window applied to provide a better view of the extracted reflexes.
    Also provided on this plot data are the windows provided from which to compute
    reflex outcomes.

    """
    muscles, n_muscles, intensities, n_intensities = get_muscles_and_intensities(section)
    plot_width, plot_height = get_plot_width_height(n_muscles, n_intensities)

    # Additional scaling to figure size added based on appearance
    plt.figure(figsize=(plot_width * 2, 2 * plot_height))
    _gen_figure(muscles, n_muscles, intensities, n_intensities, section)
    plt.tight_layout()
    save_figure(section)
    plt.close()


def get_muscles_and_intensities(section):
    muscles = list(section.reflexes.keys())
    n_muscles = len(muscles)
    intensities = list(section.reflexes[muscles[0]].avg_waveform.keys())
    n_intensities = 1
    if intensities is not None:
        n_intensities = len(intensities)
    return muscles, n_muscles, intensities, n_intensities


def get_plot_width_height(n_muscles, n_intensities):
    plot_width = PLOT_WIDTH_SCALING * n_muscles
    plot_height = PLOT_HEIGHT_SCALING * n_intensities
    return plot_width, plot_height


def _gen_figure(muscles, n_muscles, intensities, n_intensities, section):
    """For each intensity of each muscle, add subplots"""
    subplot = 1
    for intensity in intensities:
        for muscle in muscles:
            # Extracted reflex
            plt.subplot(n_intensities, n_muscles * 2, subplot)
            overlay_reflexes(section, muscle, intensity)

            # Same reflex but with adjusted x-axis and reflex windows highlighted
            plt.subplot(n_intensities, n_muscles * 2, subplot + 1)
            overlay_reflexes(section, muscle, intensity, legend=False)
            set_x_lim(section.info.windows.ms_plotting.single)
            add_reflex_windows(section.info, muscle)
            subplot += 2  # Iterate by 2 to skip over zoomed version of previous muscle


def overlay_reflexes(section, muscle, intensity, legend=True):
    """Plot overlayed raw reflexes and their average

    `intensity` is a numerical value or None.
    `legend` can be turned off for subplots where we want to show reflex windows.

    """
    # First, plot the overlayed raw reflexes
    count = 0
    for reflex in section.reflexes[muscle].reflexes:
        if intensity is None:
            if legend:
                plt.plot(section.reflexes[muscle].x_axis_extract,
                         reflex.waveform,
                         label=(f'{reflex.extract_times[0]:6.1f}-'
                                f'{reflex.extract_times[1]:6.1f}'),
                         alpha=0.8)
            else:
                plt.plot(section.reflexes[muscle].x_axis_extract,
                         reflex.waveform,
                         alpha=0.8)
            count += 1
        else:
            if intensity == reflex.stim_intensity:
                if legend:
                    plt.plot(section.reflexes[muscle].x_axis_extract,
                             reflex.waveform,
                             label=(f'{reflex.extract_times[0]:6.1f}-'
                                    f'{reflex.extract_times[1]:6.1f}'),
                             alpha=0.8)
                    count += 1
                else:
                    plt.plot(section.reflexes[muscle].x_axis_extract,
                             reflex.waveform,
                             alpha=0.8)
                    count += 1
    # Second, plot the average of the raw reflexes
    if intensity is None:
        if section.reflexes[muscle].avg_waveform is not None:
            for _, value in section.reflexes[muscle].avg_waveform.items():
                # If `single` processed, value will be a `Single` instance
                # If `train` processed, value will be ndarray
                if isinstance(value, numpy.ndarray):
                    plt.plot(section.reflexes[muscle].x_axis_extract,
                             value,
                             alpha=1,
                             color='black',
                             linewidth=0.5)
                else:
                    plt.plot(section.reflexes[muscle].x_axis_extract,
                             value.waveform,
                             alpha=1,
                             color='black',
                             linewidth=0.5)

    # If there are reflexes collected at different stimulation intensities
    else:
        if section.reflexes[muscle].avg_waveform is not None:
            try:
                avg_reflex = section.reflexes[muscle].avg_waveform[intensity]
                if isinstance(avg_reflex, numpy.ndarray):
                    plt.plot(section.reflexes[muscle].x_axis_extract,
                             avg_reflex,
                             alpha=1,
                             color='black',
                             linewidth=0.5)
                else:
                    plt.plot(section.reflexes[muscle].x_axis_extract,
                             avg_reflex.waveform,
                             alpha=1,
                             color='black',
                             linewidth=0.5)
            except KeyError:
                pass

    plt.xlabel('time (s)')
    plt.grid()
    if intensity is None:
        plt.title((f'{section.info.subject}-'
                   f'{section.info.trial}-'
                   f'{section.info.section}-'
                   f'{muscle}'), fontsize=10)
    else:
        plt.title((f'{section.info.subject}-'
                   f'{section.info.trial}-'
                   f'{section.info.section}-'
                   f'{muscle}-'
                   f'{intensity}mA'), fontsize=10)
    if (count < 11) and legend:
        plt.legend(fontsize='x-small')


def add_reflex_windows(info, muscle):
    """Add transparent colored box indicating window for reflex outcome calculations"""
    axes = plt.gca()
    y_lim = axes.get_ylim()
    if info.windows.ms_reflexes is None:
        return
    legend_items = list()
    legend_names = list()
    try:
        # Add all specified reflex windows
        for section_name, section_reflex_info in info.windows.ms_reflexes[muscle].items():
            if section_name == info.section:
                for i, (reflex, reflex_window) in enumerate(section_reflex_info.items()):
                    xy = [reflex_window[0] * s2pr.utils.CONVERT_MS_TO_S, y_lim[0]]
                    width = (reflex_window[1] * s2pr.utils.CONVERT_MS_TO_S) - (reflex_window[0] * s2pr.utils.CONVERT_MS_TO_S)
                    height = y_lim[1] - y_lim[0]
                    rect = patches.Rectangle(xy=xy,
                                             width=width,
                                             height=height,
                                             linewidth=1,
                                             edgecolor=REFLEX_COLORS[i],
                                             facecolor=REFLEX_COLORS[i],
                                             alpha=0.3,
                                             label=reflex)
                    axes.add_patch(rect)
                    legend_items.append(rect)
                    legend_items.append(reflex)
        plt.legend(handles=legend_items, labels=legend_names, fontsize='x-small')
    except KeyError:
        print('key error')


def set_x_lim(xlim):
    plt.xlim([xlim[0] * s2pr.utils.CONVERT_MS_TO_S, xlim[1] * s2pr.utils.CONVERT_MS_TO_S])


def save_figure(section):
    plt.tight_layout()
    figure_name = f'{section.info.trial}_{section.info.section}_reflexes.pdf'
    path = section.info.study_path / section.info.subject / 'figures' / 'reflexes'
    if not path.exists():
        path.mkdir()
    plt.savefig(path/figure_name)



import spike2py_reflex as s2pr


def test_init(study1_path):
    info = s2pr.info.Info(study1_path)
    assert info.subjects == ["sub01"]
    assert info.trial == None
    assert info.study_name == 'TSS_H-reflex'


def test_init_subject_section_with_no_reflex_file(study1_path):
    info = s2pr.info.Info(study1_path)
    info.init_subject('sub01')
    info.trial = 'biphasic_high_fq'
    info.init_section('ramp')
    assert info.subject == 'sub01'
    assert info.section == 'ramp'
    assert info.stim_params.train_fq == 21
    assert info.windows._section.ms.sd == [-205, -5]
    expected = {"biphasic_high_fq": ["mmax", "ramp", "hreflex"]}
    assert info.trials_sections.trials_sections == expected


def test_init_subject_section_with_reflex_file(study1_path):
    info = s2pr.info.Info(study1_path)
    info.init_subject('sub01')
    info.trial = 'biphasic_high_fq'
    info.init_section('mmax')
    assert info.subject == 'sub01'
    assert info.section == 'mmax'
    assert info.stim_params.train_fq == 66
    assert info.windows._section.ms.sd == [-102, -7]
    expected = {"biphasic_high_fq": ["mmax", "ramp", "hreflex"]}
    assert info.trials_sections.trials_sections == expected


def test_init_subject_section_with_reflex_file_and_not(study1_path):
    info = s2pr.info.Info(study1_path)
    info.init_subject('sub01')
    info.trial = 'biphasic_high_fq'
    info.init_section('mmax')
    assert info.subject == 'sub01'
    assert info.section == 'mmax'
    assert info.stim_params.train_fq == 66
    assert info.windows._section.ms.sd == [-102, -7]
    info.clear_section()
    info.init_section('ramp')
    assert info.stim_params.train_fq == 21
    assert info.windows._section.ms.sd == [-205, -5]



"""Class to hold channel-related info that updates as required."""

from typing import Union
from dataclasses import dataclass


@dataclass
class ChannelsInfo:
    emg: list
    triggers: dict
    stim_intensity: Union[str, None] = None


class Channels:
    """Channel information to process study, subject and section.

    Channel information is stored in an instance of `ChannelsInfo`, but is
    externally accessed as class properties:
      - Channels.emg
      - Channels.triggers
      - Channels.stim_intensity
    By accessing the data in this way when process a given section, the user
    will be sure to be accessing the correct information. section-level
    information trumps subject-level, which trumps study-level.
    """

    def __init__(self, study_channel_info: Union[None, dict] = None):
        """
        Parameters
        ----------
        study_channel_info
            Contains information about channels required to process data with
            spike2py_reflex. Expected keys are "emg", "triggers" and
            "stim_intensity". The values associated with the "triggers" key
            are the name of relevant spike2py_reflex section, the spike2py
            channel name associated with the trigger, and the type of trigger
            ("single", "double", "train"). "stim_intensity" has a value
            corresponding to the spike2py channel name containing the stimulus
            intensity (if one was used).

        Examples
        --------
        {"emg": ["Fdi"],
        "triggers": {
            "mmax": {
                "channels": "Mmax",
                "type": "single"
                },
            "hreflex": {
                "channels": "Ds8",
                "type": "double"
                }
            },
        "stim_intensity": "Stim"
        }

        """
        self._study = None
        self._subject = None
        self._section = None
        if study_channel_info is not None:
            self._study = ChannelsInfo(**study_channel_info)

    def add_subject(self, subject_channel_info=None):
        """Add channel information for a subject"""
        if subject_channel_info is not None:
            self._subject = ChannelsInfo(**subject_channel_info)
        elif subject_channel_info is None:
            self._subject = self._study

    def add_section(self, section_channel_info=None):
        """Add channel information for a section"""
        if section_channel_info is not None:
            self._section = ChannelsInfo(**section_channel_info)
        elif section_channel_info is None:
            self._section = self._subject

    def clear_subject(self):
        """Clear subject-level channel information"""
        self._subject = None

    def clear_section(self):
        """Clear section-level channel information"""
        self._section = None

    @property
    def emg(self) -> list:
        """List of one or more emg channels from which to extract reflexes"""
        return self._section.emg

    @property
    def triggers(self) -> dict:
        """Dict for each section specifying trigger channel and trigger type"""
        return self._section.triggers

    @property
    def stim_intensity(self) -> Union[str, None]:
        """Str of specifying stimulation intensity channel (if used)"""
        return self._section.stim_intensity



import spike2py_reflex as s2pr


def extract(info, data):
    """Extract reflexes for all muscles (emg) in section.

    Parameters
    ----------
    info: s2pr.Info
      Contains details about all aspects of the study and its processing
    data: spike2py.trial.Trial
      Data of trial section being analysed

    Returns
    -------
    s2pr.SectionReflexes, which contains the reflexes and supplementary details
    of the current setion.
    """
    stim_intensities = s2pr.utils.get_stim_intensity(info, data)
    extracted = dict()

    for emg_name in info.channels.emg:
        emg = getattr(data, emg_name)
        muscle_reflexes = dict()
        # specify sampling frequency to trigger computation of idx
        info.windows.fs = emg.info.sampling_frequency
        if info.triggers.type in [s2pr.utils.SINGLE, s2pr.utils.TRAIN]:
            muscle_reflexes = _extract_single_reflexes(
                emg_name, emg, stim_intensities, info
            )
        elif info.triggers.type == s2pr.utils.DOUBLE:
            muscle_reflexes = _extract_double_reflexes(
                emg_name, emg, stim_intensities, info
            )
        extracted[emg_name] = muscle_reflexes

    return s2pr.reflexes.SectionReflexes(info, extracted)


def _extract_single_reflexes(emg_name, emg, stim_intensities, info):
    """Extract reflexes from single or train stimulation"""

    extract_idxs = _get_extract_idx_singles(info)
    x_axis = _get_x_axis(info)
    reflexes = _get_single_reflexes(extract_idxs, stim_intensities, emg)

    muscle_reflexes = s2pr.reflexes.Singles(
        x_axis_extract=x_axis,
        reflexes=reflexes,
        type=info.triggers.type,
        sd_window_idx=info.windows.idx_sd,
        sd_window_ms=info.windows.ms_sd,
        reflex_windows_idx=info.windows.idx_reflexes[emg_name],
        reflex_windows_ms=info.windows.ms_reflexes[emg_name],
    )
    return muscle_reflexes


def _get_extract_idx_singles(info):
    """Get idx of each window from which to extract single reflexes"""
    extract_idxs = list()
    trigger_type = s2pr.utils.SINGLE
    if info.triggers.type == s2pr.utils.TRAIN:
        trigger_type = s2pr.utils.TRAIN_SINGLE_PULSE
    extract_idx = getattr(info.windows.idx_extract, trigger_type)
    for trigger_idx in info.triggers.extract:
        extract_idxs.append(_get_window(trigger_idx, extract_idx))
    if _start_of_first_window_out_of_bounds(extract_idxs):
        extract_idxs.pop(0)
    return extract_idxs


def _get_window(trigger_idx, window_idx):
    """Get a pair of idx values to extract reflex from a trigger"""
    if trigger_idx is None:
        return [None, None]
    # TODO: Presumably window_idx[0] is negative? If so, make note here
    lower = trigger_idx + window_idx[0]
    upper = trigger_idx + window_idx[1]
    return [lower, upper]


def _start_of_first_window_out_of_bounds(extract_idxs):
    return extract_idxs[0][0] < 0


def _get_x_axis(info):
    x_axis = list()
    if info.triggers.type == s2pr.utils.SINGLE:
        x_axis = info.windows.x_axes.single
    elif info.triggers.type == s2pr.utils.TRAIN:
        x_axis = info.windows.x_axes.train_single_pulse
    return x_axis


def _get_single_reflexes(extract_idxs: list, stim_intensities: list, emg):
    """
    Extract reflexes based on window idx for each trigger

    Parameters
    ----------
    extract_idxs: pairs of indexes for each reflex to extract
    stim_intensities: Intensities associated with each reflex (if stim intensity recorded)
    emg: spike2py data for given emg channel from which to extract reflexes

    Returns
    -------
    all extracted reflexes

    """
    reflexes = list()
    for (idx1_extract, idx2_extract), intensity in zip(
            extract_idxs, stim_intensities
    ):
        try:
            reflexes.append(
                s2pr.reflexes.Single(
                    waveform=emg.values[idx1_extract:idx2_extract],
                    extract_times=(emg.times[idx1_extract], emg.times[idx2_extract]),
                    extract_indexes=(idx1_extract, idx2_extract),
                    stim_intensity=intensity,
                )
            )
        except IndexError:
            print(f'Dropped a reflex asking for idx1 {idx1_extract} and idx2 {idx2_extract},'
                  f'from signal with length {len(emg.times)}')
    return reflexes


def _extract_double_reflexes(emg_name, emg, stim_intensities, info):
    """Extract reflexes from doubles"""

    trigger_windows = _get_extract_idx_doubles(info)
    reflexes = _get_double_reflexes(info, trigger_windows, stim_intensities, emg)

    muscle_reflexes = s2pr.reflexes.Doubles(
        x_axis_extract=info.windows.x_axes.double,
        x_axis_singles=info.windows.x_axes.double_single_pulse,
        reflexes=reflexes,
        type=info.triggers.type,
        sd_window_idx=info.windows.idx_sd,
        sd_window_ms=info.windows.ms_sd,
        reflex_windows_idx=info.windows.idx_reflexes[emg_name],
        reflex_windows_ms=info.windows.ms_reflexes[emg_name],
    )
    return muscle_reflexes


def _get_extract_idx_doubles(info) -> list:
    """Get windows to extract double, and each reflex individually

    Returns
    -------

    List of lists. Each item in list represents one double. For each of these, there are three pairs of indexes.
    These are to extract 1) the entire double, 2) the first reflex, and 3) the second reflex.

    e.g. [[extract_start, extract_end], [reflex1_start, reflex1_end], [reflex2_start, reflex2_end]] , ...]
    """
    extract_idxs = list()
    extract_idx = getattr(info.windows.idx_extract, info.triggers.type)
    double_idx = getattr(info.windows.idx_extract, "double_single_pulse")
    for trigger_idx_extract, trigger_idx_double in zip(
        info.triggers.extract, info.triggers.double
    ):
        extract = _get_window(trigger_idx_extract, extract_idx)
        reflex1 = _get_window(trigger_idx_double[0], double_idx)
        reflex2 = _get_window(trigger_idx_double[1], double_idx)
        extract_idxs.append([extract, reflex1, reflex2])
    return extract_idxs


def _get_double_reflexes(info, trigger_windows, stim_intensities, emg):
    reflexes = list()
    stim_intensities_second_pulse_doubles_removed = _remove_second_pulse(
        stim_intensities,
        info.triggers._triggers,
        info.windows.double_isi
    )
    for (
            (idx1_extract, idx2_extract),
            (idx1_reflex1, idx2_reflex1),
            (idx1_reflex2, idx2_reflex2),
    ), (intensity) in zip(trigger_windows, stim_intensities_second_pulse_doubles_removed):

        reflex1 = s2pr.reflexes.Single(waveform=emg.values[idx1_reflex1:idx2_reflex1],
                                       extract_indexes=(idx1_reflex1, idx2_reflex1),
                                       extract_times=(emg.times[idx1_reflex1], emg.times[idx2_reflex1]),
                                       )

        reflex2 = None
        if idx1_reflex2 is not None:
            reflex2 = s2pr.reflexes.Single(
                waveform=emg.values[idx1_reflex2:idx2_reflex2],
                extract_indexes=(idx1_reflex2, idx2_reflex2),
                extract_times=(emg.times[idx1_reflex2], emg.times[idx2_reflex2])
            )
        double = s2pr.reflexes.Double(
            waveform=emg.values[idx1_extract:idx2_extract],
            reflex1=reflex1,
            reflex2=reflex2,
            stim_intensity=intensity,
            extract_indexes=(idx1_extract, idx2_extract),
            extract_times=(emg.times[idx1_extract], emg.times[idx2_extract])
        )
        reflexes.append(double)

    return reflexes


# TODO: Add test that ensures that when doubles given, only first stim intensity retained
# TODO: Add test that ensures that when mix of singles and doubles, only first stim of double retained, plus all singles
def _remove_second_pulse(stim_intensities, triggers, double_isi):
    stim_intensities_second_pulse_doubles_removed = [stim_intensities[0]]
    for i in range(1, len(triggers)):
        isi = triggers[i] - triggers[i-1]
        if _pulse_not_second_in_double(isi, double_isi):
            stim_intensities_second_pulse_doubles_removed.append(stim_intensities[i])
    return stim_intensities_second_pulse_doubles_removed


def _pulse_not_second_in_double(isi, double_isi):
    return isi > (double_isi*s2pr.utils.CONVERT_MS_TO_S) * 1.05



from pathlib import Path

import spike2py_reflex as s2pr


def test_process_study():
    study_path = Path('/home/martin/Dropbox/Martin/sketchbook/python/projects/spike2py_reflex/test/data/study1')
    s2pr.process.study(study_path, plot=True)


def test_process_subject():
    study_path = Path('/home/martin/Dropbox/Martin/sketchbook/python/projects/spike2py_reflex/test/data/study1')
    subject_ = 'sub01'
    from_command_line = True
    s2pr.process.subject(subject_=subject_,
                         study_path=study_path,
                         plot=True,
                         from_command_line=from_command_line)


def test_process_trial():
    study_path = Path('/home/martin/Dropbox/Martin/sketchbook/python/projects/spike2py_reflex/test/data/study1')
    subject_ = 'sub01'
    from_command_line = True
    trial = 'biphasic_high_fq'
    section_ = ["mmax", "ramp", "hreflex"]
    s2pr.process.trial(trial_=trial,
                         sections=section_,
                         subject_=subject_,
                         study_path=study_path,
                         plot=True,
                         from_command_line=from_command_line)



"""Package with classes and dataclasses to manage spike2py_reflex information

The main class is `Info`. It is populated and updated as data is being
processed. It holds and updates instances of the other classes: Channel,
StimParams, TrialSections, and Windows.

A key functionality of `Info` and the other classes is that, depending on
whether the subject or the trial required tweaks to aspects of the info, the
changes to the info will be retained for given trial(s) and subject(s). In all
other instances the study-level info will be used.

"""

from .channels import Channels
from .stim_params import StimParams
from .windows import Window, Windows, WindowTypes, GroupedWindows
from .trials_sections import TrialsSections
from .info import Info



import spike2py_reflex as s2pr
import pytest


def test_reflex_outcomes_hreflex_doubles_individual_reflexes(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    assert section.reflexes['Fdi'].reflexes[0].reflex1.outcomes['hreflex'].onset == pytest.approx(0.015625)
    assert section.reflexes['Fdi'].reflexes[0].reflex2.outcomes['hreflex'].peak_to_peak == pytest.approx(-0.005965828180407999)


def test_reflex_outcomes_ramp_doubles_individual_reflexes(info_data_ramp):
    info, data = info_data_ramp
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    assert section.reflexes['Fdi'].reflexes[0].outcomes['cool_reflex'].peak_to_peak == pytest.approx(-0.005139612876265053)
    assert section.reflexes['Fdi'].reflexes[0].outcomes['cool_reflex'].area == pytest.approx(0.15115431510702879)
    assert section.reflexes['Fdi'].reflexes[0].outcomes['cool_reflex'].onset is None


def test_reflex_outcomes_mmax_doubles_individual_reflexes(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    print(section.reflexes['Fdi'].reflexes[0].outcomes)
    assert section.reflexes['Fdi'].reflexes[0].outcomes['mmax'].peak_to_peak == pytest.approx(-0.0036865157257219843)
    assert section.reflexes['Fdi'].reflexes[0].outcomes['mmax'].area == pytest.approx(67.80426369321032)
    assert section.reflexes['Fdi'].reflexes[0].outcomes['hreflex'].onset == pytest.approx(0.020161290322580645)


def test_reflex_outcomes_hreflex_doubles_avg(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    section = s2pr.outcomes.calculate_outcomes_of_avg(section)
    assert len(section.reflexes['Fdi'].avg_reflex1[19].waveform) == 109
    assert section.reflexes['Fdi'].avg_reflex1[19].outcomes[
               'hreflex'].peak_to_peak == pytest.approx(-0.0003608836017132724)
    assert section.reflexes['Fdi'].avg_reflex2[19].outcomes[
               'hreflex'].area == pytest.approx(0.06735172606086091)


def test_reflex_outcomes_ramp_singles_avg(info_data_ramp):
    info, data = info_data_ramp
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    section = s2pr.outcomes.calculate_outcomes_of_avg(section)
    assert len(section.reflexes['Fdi'].avg_waveform) == 3
    assert len(section.reflexes['Fdi'].avg_waveform[10].waveform) == 90
    assert len(section.reflexes['Fdi'].avg_waveform[24].waveform) == 90
    assert section.reflexes['Fdi'].avg_waveform[10].outcomes['cool_reflex'].peak_to_peak == pytest.approx(0.012636357578906982)
    assert section.reflexes['Fdi'].avg_waveform[10].outcomes[
               'cool_reflex'].area == pytest.approx(0.06081335857846465)



def test_reflex_outcomes_mmax_singles_avg(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    section = s2pr.outcomes.calculate_outcomes_of_avg(section)
    assert len(section.reflexes['Fdi'].avg_waveform) == 1
    assert len(section.reflexes['Fdi'].avg_waveform[1].waveform) == 694
    assert section.reflexes['Fdi'].avg_waveform[1].outcomes['mmax'].peak_to_peak == pytest.approx(0.000937179558910316)
    assert section.reflexes['Fdi'].avg_waveform[1].outcomes[
               'hreflex'].peak_to_peak == pytest.approx(0.001988228210304802)


def test_reflex_outcomes_hreflex_doubles_mean(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    section = s2pr.outcomes.calculate_mean_outcomes(section)
    assert section.reflexes['Fdi'].mean_outcomes_reflex1['hreflex'][19]['outcomes'].peak_to_peak == pytest.approx(-0.00013157086170218784)
    assert section.reflexes['Fdi'].mean_outcomes_reflex1['hreflex'][19]['missing_outcomes'].peak_to_peak == 0
    assert section.reflexes['Fdi'].mean_outcomes_reflex1['hreflex'][19][
        'missing_outcomes'].onset == 0
    assert section.reflexes['Fdi'].mean_outcomes_reflex1['hreflex'][19][
        'present_outcomes'].onset == 7
    assert section.reflexes['Fdi'].mean_ratio['hreflex'][19]['ratio'] == pytest.approx(-1.0970691361542422)
    assert section.reflexes['Fdi'].mean_ratio['hreflex'][19]['missing_ratio'] == 0
    assert section.reflexes['Fdi'].mean_ratio['hreflex'][19]['present_ratio'] == 7


def test_reflex_outcomes_ramp_train_mean(info_data_ramp):
    info, data = info_data_ramp
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    section = s2pr.outcomes.calculate_mean_outcomes(section)
    assert section.reflexes['Fdi'].mean_outcomes['cool_reflex'][10]['outcomes'].peak_to_peak == pytest.approx(0.0010047763859973992)
    assert section.reflexes['Fdi'].mean_outcomes['cool_reflex'][10]['missing_outcomes'].peak_to_peak == 0
    assert section.reflexes['Fdi'].mean_outcomes['cool_reflex'][10]['missing_outcomes'].onset == 19
    assert section.reflexes['Fdi'].mean_outcomes['cool_reflex'][24][
               'missing_outcomes'].peak_to_peak == 0


def test_reflex_outcomes_mmax_single_mean(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    section = s2pr.outcomes.calculate_mean_outcomes(section)
    assert section.reflexes['Fdi'].mean_outcomes['mmax'][1]['outcomes'].peak_to_peak == pytest.approx(-0.0001343032256506074)
    assert section.reflexes['Fdi'].mean_outcomes['hreflex'][1]['missing_outcomes'].peak_to_peak == 0


def test_reflex_outcomes_mmax_single_mean_second_muscle(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section = s2pr.outcomes.calculate_for_individual_reflexes(section)
    section = s2pr.outcomes.calculate_mean_outcomes(section)
    assert section.reflexes['Adm'].mean_outcomes['mmax'][1]['outcomes'].peak_to_peak == pytest.approx(-0.0001343032256506074)
    assert section.reflexes['Adm'].mean_outcomes['hreflex'][1]['missing_outcomes'].peak_to_peak == 0



import spike2py_reflex as s2pr


def test_init():
    sp = s2pr.info.StimParams()
    assert sp._study is None
    assert sp._subject is None
    assert sp._section is None


def test_get_attributes_simple(demo_study_reflex):
    study_sp_info = demo_study_reflex['stim_params']
    sp = s2pr.info.StimParams(study_sp_info)
    sp.add_subject()
    sp.add_section()
    assert sp.train_fq == 30
    assert sp.double_isi == 50
#    assert sp.kHz_fq == 10


def test_get_attributes_complex1(demo_study_reflex):
    study_sp_info = demo_study_reflex['stim_params']
    sp = s2pr.info.StimParams(study_sp_info)
    sub_sp = {"train_fq": 10, "double_isi": 75}

    sp.add_subject(sub_sp)
    sp.add_section()
    assert sp.train_fq == 10
    assert sp.double_isi == 75


def test_get_attributes_complex2(demo_study_reflex):
    study_sp_info = demo_study_reflex['stim_params']
    sp = s2pr.info.StimParams(study_sp_info)
    sp.add_subject()
    sec_sp = {"train_fq": 40, "double_isi": 25}
    sp.add_section(sec_sp)
    assert sp.train_fq == 40
    assert sp.double_isi == 25



import spike2py_reflex as s2pr


def test_get_stim_intensity(info_data_hreflex):
    info, data = info_data_hreflex
    intensities = s2pr.utils.get_stim_intensity(info, data)
    assert intensities == [19, 19, 19, 19, 19, 19, 19]



from dataclasses import dataclass


@dataclass
class StimParamsInfo:
#    kHz_fq: float
    train_fq: int
    double_isi: float


class StimParams:
    def __init__(self, study_stim_params_info=None):
        self._study = None
        self._subject = None
        self._section = None
        if study_stim_params_info is not None:
            self._study = StimParamsInfo(**study_stim_params_info)

    def add_subject(self, subject_stim_params_info=None):
        if subject_stim_params_info is not None:
            self._subject = StimParamsInfo(**subject_stim_params_info)
        elif subject_stim_params_info is None:
            self._subject = self._study

    def add_section(self, section_stim_params_info=None):
        if section_stim_params_info is not None:
            self._section = StimParamsInfo(**section_stim_params_info)
        elif section_stim_params_info is None:
            self._section = self._subject

    def clear_subject(self):
        self._subject = None

    def clear_section(self):
        self._section = None

    @property
    def kHz_fq(self):
        if self._section is not None:
            return self._section.kHz_fq
        elif self._subject is not None:
            return self._subject.kHz_fq
        elif self._study is not None:
            return self._study.kHz_fq
        else:
            return None

    @property
    def train_fq(self):
        if self._section is not None:
            return self._section.train_fq
        elif self._subject is not None:
            return self._subject.train_fq
        elif self._study is not None:
            return self._study.train_fq
        else:
            return None

    @property
    def double_isi(self):
        if self._section is not None:
            return self._section.double_isi
        elif self._subject is not None:
            return self._subject.double_isi
        elif self._study is not None:
            return self._study.double_isi
        else:
            return None



import spike2py_reflex as s2pr


def test_triggers_singles(info_data_mmax):
    triggers = s2pr.utils.Triggers(info_data_mmax[0], info_data_mmax[1])
    assert triggers.type == 'single'
    print(triggers._raw)
    assert round(triggers._raw[0]) == 155
    assert len(triggers._raw) == 6
    assert triggers.extract == [17453, 62332, 98842, 139727, 178267, 199410]
    assert len(triggers.extract) == 6
    assert triggers.double is None


def test_triggers_doubles(info_data_hreflex):
    triggers = s2pr.utils.Triggers(info_data_hreflex[0], info_data_hreflex[1])
    assert triggers.type == 'double'
    assert round(triggers._raw[0]) == 806
    assert len(triggers._raw) == 14
    assert triggers.extract == [18775, 54235, 80455, 100836, 119376, 140202, 159884]
    assert triggers.double == [[18775, 18874],
                               [54235, 54334],
                               [80455, 80555],
                               [100836, 100936],
                               [119376, 119475],
                               [140202, 140301],
                               [159884, 159983]]


def test_triggers_ramp(info_data_ramp):
    triggers = s2pr.utils.Triggers(info_data_ramp[0], info_data_ramp[1])
    assert triggers.type == 'train'
    assert round(triggers._raw[0]) == 1504
    assert len(triggers._raw) == 1663
    assert len(triggers.extract) == 1663
    assert triggers.double is None





import spike2py_reflex as s2pr


def test_init():
    chan = s2pr.info.Channels()
    assert chan._study is None
    assert chan._subject is None
    assert chan._section is None


def test_add_channels(demo_study_reflex):
    study_channel_info = demo_study_reflex['channels']
    chan = s2pr.info.Channels(study_channel_info)
    assert chan._study.emg == ['Fdi']
    assert chan._study.stim_intensity == 'Stim'
    expected = {'hreflex': {'channels': 'Ds8', 'type': 'double'},
 'mmax': {'channels': 'Mmax', 'type': 'single'},
 'ramp': {'channels': 'Ds8', 'type': 'single'}}
    assert chan._study.triggers == expected


def test_add_subject_no_channels(demo_study_reflex):
    study_channel_info = demo_study_reflex['channels']
    chan = s2pr.info.Channels(study_channel_info)
    chan.add_subject()
    assert chan._subject.emg == ['Fdi']
    assert chan._subject.stim_intensity == 'Stim'
    expected = {'hreflex': {'channels': 'Ds8', 'type': 'double'},
                'mmax': {'channels': 'Mmax', 'type': 'single'},
                'ramp': {'channels': 'Ds8', 'type': 'single'}}
    assert chan._subject.triggers == expected


def test_add_subject_channels(demo_study_reflex, subject_channels):
    study_channel_info = demo_study_reflex['channels']
    chan = s2pr.info.Channels(study_channel_info)
    chan.add_subject(subject_channels)
    assert chan._study.emg == ['Fdi']
    assert chan._subject.emg == ['MG']
    assert chan._study.stim_intensity == 'Stim'
    assert chan._subject.stim_intensity == 'zap'


def test_add_section_no_channels(demo_study_reflex):
    study_channel_info = demo_study_reflex['channels']
    chan = s2pr.info.Channels(study_channel_info)
    chan.add_subject()
    chan.add_section()
    assert chan._section.emg == ['Fdi']
    assert chan._section.stim_intensity == 'Stim'
    expected = {'hreflex': {'channels': 'Ds8', 'type': 'double'},
                'mmax': {'channels': 'Mmax', 'type': 'single'},
                'ramp': {'channels': 'Ds8', 'type': 'single'}}
    assert chan._section.triggers == expected


def test_add_section_channels(demo_study_reflex, subject_channels, section_channels):
    study_channel_info = demo_study_reflex['channels']
    chan = s2pr.info.Channels(study_channel_info)
    chan.add_subject(subject_channels)
    chan.add_section(section_channels)
    assert chan._section.emg == ['hh']
    assert chan._section.stim_intensity == 'zapzap'
    expected = {'H_reflex': {'channels': 'trig3', 'type': 'double'},
 'Mmax': {'channels': 'trig1', 'type': 'double'},
 'rRamp': {'channels': 'trig2', 'type': 'single'}}
    assert chan._section.triggers == expected

def test_clear(demo_study_reflex, subject_channels, section_channels):
    study_channel_info = demo_study_reflex['channels']
    chan = s2pr.info.Channels(study_channel_info)
    chan.add_subject(subject_channels)
    chan.add_section(section_channels)
    chan.clear_section()
    assert chan._section == None
    chan.clear_subject()
    assert chan._subject == None


def test_get_attributes_simple(demo_study_reflex):
    study_channel_info = demo_study_reflex['channels']
    chan = s2pr.info.Channels(study_channel_info)
    chan.add_subject()
    chan.add_section()
    assert chan.emg == ['Fdi']
    assert chan.stim_intensity == 'Stim'
    expected = {'hreflex': {'channels': 'Ds8', 'type': 'double'},
                'mmax': {'channels': 'Mmax', 'type': 'single'},
                'ramp': {'channels': 'Ds8', 'type': 'single'}}
    assert chan.triggers == expected


def test_get_attributes_complex(demo_study_reflex, subject_channels, section_channels):
    study_channel_info = demo_study_reflex['channels']
    chan = s2pr.info.Channels(study_channel_info)
    chan.add_subject(subject_channels)
    chan.add_section(section_channels)
    assert chan.emg == ['hh']
    assert chan.stim_intensity == 'zapzap'
    expected = {'H_reflex': {'channels': 'trig3', 'type': 'double'},
 'Mmax': {'channels': 'trig1', 'type': 'double'},
 'rRamp': {'channels': 'trig2', 'type': 'single'}}
    assert chan.triggers == expected



import sys

from spike2py_preprocess.trial_sections import find_nearest_time_index
import spike2py_reflex as s2pr


def get_stim_intensity(info, data) -> list:
    ch = _get_stim_intensity_channel(info, data)
    intensities = list()
    for trigger in info.triggers._triggers:
        intensity_value = None
        if ch is not None:
            idx = find_nearest_time_index(ch.times, trigger)
            intensity_value = round(ch.values[idx])
        intensities.append(intensity_value)
    try:
        intensities = _verify_doubles_at_same_intensities(info, intensities, info.windows.double_isi)
    except AttributeError:
        pass
    return intensities


# TODO: add test where there is different stim intensities (Janie H-reflex trial S6 '1000us_bi_h')
# TODO: add test where there is different stim intensities for a double (Janie TSS trial S7 '400us_bi_tss_tss')


def _verify_doubles_at_same_intensities(info, intensities, double_isi):
    triggers = info.triggers._triggers
    for i, (trigger, intensity) in enumerate(zip(triggers, intensities)):
        if i == 0:
            previous_trigger_time = trigger
            previous_trigger_intensity = intensity
        current_trigger_time = trigger
        current_trigger_intensity = intensity
        isi = current_trigger_time - previous_trigger_time
        if isi < (double_isi*s2pr.utils.CONVERT_MS_TO_S) * 1.05:
            if previous_trigger_intensity != current_trigger_intensity:
                print('!' * 80,
                      f'\nIntensity of first and second trigger for double not equal.\n\n'
                      f'\t\tsubject: {info.subject}\n'
                      f'\t\ttrial: {info.trial}\n'
                      f'\t\tsection: {info.section}\n',
                      f'\t\ttrigger 1 time:{previous_trigger_time}\n',
                      f'\t\tstim intensity 1:{previous_trigger_intensity}\n',
                      f'\t\ttrigger 2 time:{current_trigger_time}\n',
                      f'\t\tstim intensity 2:{current_trigger_intensity}\n\n',
                      'To remove this double from the section, please add the window'
                      f'[{previous_trigger_time-0.2}, {current_trigger_time + 0.2}]\n'
                      f'to the subject trial json file under "rejected_trigger_windows"\n',
                      '!' * 80,
                      '\n')
                sys.exit(1)
    return intensities


def _get_stim_intensity_channel(info, data):
    stim_intensity = None
    try:
        stim_intensity = getattr(data, info.channels.stim_intensity)
    except AttributeError:
        print(f'\t\tstudy: {info.study}; '
              f'trial: {info.trial}; '
              f'section: {info.section} --> no `stim_intensity` channel.')
    return stim_intensity



import matplotlib.pyplot as plt

import spike2py_reflex as s2pr

PLOT_WIDTH_SCALING = 1
PLOT_HEIGHT_SCALING = 1


def plot_train(section):
    """Plot section train reflexes (singles).

    Generates a plot with n columns, where n = number_muscles, and m rows,
    where m = number of stimulation intensities. Each column plots the data
    extracted for each reflex, and also the avg reflex, and this with the `plot`
    window applied to provide a better view of the extracted reflexes. Also
    provided on this plot data are the windows provided from which to compute
    reflex outcomes.

    """

    info = section.info
    muscles, n_muscles, intensities, n_intensities = s2pr.plot.get_muscles_and_intensities(section)
    plot_width, plot_height = s2pr.plot.get_plot_width_height(n_muscles, n_intensities)
    plt.figure(figsize=(plot_width, plot_height * 1))
    _gen_figure(muscles, n_muscles, intensities, n_intensities, section, info)
    plt.tight_layout()
    s2pr.plot.save_figure(section)
    plt.close()


def _gen_figure(muscles, n_muscles, intensities, n_intensities, section, info):
    """For each intensity of each muscle, add subplots"""
    subplot = 1
    for intensity in intensities:
        for muscle in muscles:
            plt.subplot(n_intensities, n_muscles, subplot)
            s2pr.plot.overlay_reflexes(section, muscle, intensity, legend=False)
            s2pr.plot.add_reflex_windows(info, muscle)
            s2pr.plot.set_x_lim(section.info.windows.ms_plotting.train_single_pulse)
            subplot += 1



from .reflex_outcomes import calculate, get_outcomes_from, calculate_for_individual_reflexes
from .outcomes_of_avg_reflex import get_stim_intensities, calculate_outcomes_of_avg
from .mean_outcomes import calculate_mean_outcomes
from .utils import Outcomes
from .summary import summary



import pytest
import spike2py_reflex as s2pr


@pytest.mark.plot
def test_plot_mmax_outcomes(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    s2pr.plot.single_outcomes(section)


@pytest.mark.plot
def test_plot_mmax_outcomes_two_muscles(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section.info.windows.ms_reflexes['Adm'] = section.info.windows.ms_reflexes[
        'Fdi']
    section.info.section = section.info.section + '_2'
    s2pr.plot.single_outcomes(section)


@pytest.mark.plot
def test_plot_mmax_outcomes_two_muscles_three_intensities(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    section.reflexes['Fdi'].avg_waveform[32] = \
    section.reflexes['Fdi'].avg_waveform[1]
    section.reflexes['Fdi'].avg_waveform[35] = \
    section.reflexes['Fdi'].avg_waveform[1]
    section.reflexes['Fdi'].reflexes[3].stim_intensity = 32
    section.reflexes['Fdi'].reflexes[2].stim_intensity = 32
    section.reflexes['Fdi'].reflexes[4].stim_intensity = 35
    section.reflexes['Fdi'].reflexes[5].stim_intensity = 35
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section.info.windows.ms_reflexes['Adm'] = section.info.windows.ms_reflexes[
        'Fdi']
    section.info.section = section.info.section + '_3'
    s2pr.plot.single_outcomes(section)


@pytest.mark.plot
def test_plot_outcomes_ramp_singles(info_data_ramp):
    info, data = info_data_ramp
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    s2pr.plot.train_outcomes(section)


@pytest.mark.plot
def test_plot_outcomes_ramp_singles_two_muscles(info_data_ramp):
    info, data = info_data_ramp
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section.info.windows.ms_reflexes['Adm'] = section.info.windows.ms_reflexes['Fdi']
    section.info.section = section.info.section + '_2'
    s2pr.plot.train_outcomes(section)


@pytest.mark.plot
def test_plot_outcomes_hreflex_doubles(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    s2pr.plot.double_outcomes(section)


@pytest.mark.plot
def test_plot_outocmes_hreflex_doubles_two_muscles(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section.info.windows.ms_reflexes['Adm'] = section.info.windows.ms_reflexes['Fdi']
    section.info.section = section.info.section + '_2'
    s2pr.plot.double_outcomes(section)



@pytest.mark.plot
def test_plot_outcomes_hreflex_doubles_two_muscles_three_intensities(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    section.reflexes['Fdi'].avg_waveform[25] = section.reflexes['Fdi'].avg_waveform[19]
    section.reflexes['Fdi'].avg_waveform[35] = section.reflexes['Fdi'].avg_waveform[19]
    section.reflexes['Fdi'].reflexes[3].stim_intensity = 25
    section.reflexes['Fdi'].reflexes[2].stim_intensity = 25
    section.reflexes['Fdi'].reflexes[4].stim_intensity = 35
    section.reflexes['Fdi'].reflexes[5].stim_intensity = 35
    section.reflexes['Fdi'].avg_reflex1[25] = section.reflexes['Fdi'].avg_reflex1[19]
    section.reflexes['Fdi'].avg_reflex1[35] = section.reflexes['Fdi'].avg_reflex1[19]
    section.reflexes['Fdi'].avg_reflex2[25] = section.reflexes['Fdi'].avg_reflex2[19]
    section.reflexes['Fdi'].avg_reflex2[35] = section.reflexes['Fdi'].avg_reflex2[19]
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section.info.windows.ms_reflexes['Adm'] = section.info.windows.ms_reflexes['Fdi']
    section.info.section = section.info.section + '_3'
    s2pr.plot.double_outcomes(section)



from dataclasses import dataclass
from typing import Union



@dataclass
class Outcomes:
    peak_to_peak: Union[float, None] = None
    area: Union[float, None] = None
    onset: Union[float, None] = None




from pathlib import Path
from typing import List

import spike2py_preprocess as s2pp
import spike2py_reflex as s2pr


class Info:
    """Information and functionality to required to process study data.

    Attributes
    ----------
    study_path : Path
        Path to study folder. Folder is to be structured as indicated in spike2py_preprocess docs.
    plot: Bool
        Flag indicating whether to generate plots.
    subjects: list
        List of subject IDs. Need to also be the name of the folder containing the subjects' data.
    channels: info.Channels
        Contains listing of relevant channels: emg, triggers, stim_intensity
    stim_params: info.StimParams
        Contains details related to transcutaneous spinal stimulation: kHz_fq, train_fq, double_isi
    trials_sections: info.TrialsSections
        Specifies trials and associated sections
    study_name: str
        Name of study, used in figures, numerical output, etc
    windows: info.Windows
    subject:
    trial:
    section:
    section_type:
    rejected_trig_windows:
    section_pkl:
    triggers:
    fs:
    """

    def __init__(self, study_path, plot=False):
        """
        Parameters
        ----------
        study_path: Path to study folder.
        plot: Flag indicating whether to generate plots
        """

        # Study
        self.study_path: Path = Path(study_path)
        self.plot: bool = plot

        study_info, reflex_info = self._get_study_info()

        self.subjects = study_info["subjects"]
        self.channels = s2pr.info.Channels(reflex_info["channels"])
        self.stim_params = s2pr.info.StimParams(reflex_info["stim_params"])
        self.trials_sections = s2pr.info.TrialsSections(reflex_info["trials_sections"])
        self.study_name: str = study_info["name"]
        self.windows = s2pr.info.Windows(reflex_info["windows"])
        self.windows.double_isi = self.stim_params.double_isi

        # Subject
        self.subject = None

        # Trial
        self.trial = None

        # Section
        self.section = None
        self.section_type = None
        self.rejected_trigger_windows: List[list] = None
        self.section_pkl = None
        self.triggers = None

    @property
    def fs(self):
        return self.fs

    @fs.setter
    def fs(self, emg_fs):
        self.windows.fs = emg_fs

    def _get_study_info(self):
        study_info = s2pp.utils.read_json(self.study_path / s2pr.utils.STUDY_INFO_FILE)
        reflex_info = s2pp.utils.read_json(self.study_path / s2pr.utils.REFLEX_FILE)
        return study_info, reflex_info

    def clear_section(self):
        self.section = None
        self.section_type = None
        self.rejected_trigger_windows = None
        self.triggers = None
        self.channels.clear_section()
        self.windows.clear_section()
        self.trials_sections.clear_section()
        self.stim_params.clear_section()

    def clear_trial(self):
        self.trial = None

    def clear_subject(self):
        self.subject = None
        self.channels.clear_subject()
        self.windows.clear_subject()
        self.trials_sections.clear_subject()
        self.stim_params.clear_subject()

    def init_subject(self, sub_id):
        """Initilise subject

        Parameters
        ----------
        sub_id: str
            subject id, which should also be the name of the subject folder

            """
        self.subject = sub_id

        subject_reflex_info = None

        try:
            path = self.study_path / sub_id / s2pr.utils.SUBJECT_REFLEX_FILE
            subject_reflex_info = s2pp.utils.read_json(path, strict=False)
        except FileNotFoundError:
            print(f'subject {sub_id} does not have a reflex_info.json file')

        if subject_reflex_info is not None:
            if "channels" in subject_reflex_info:
                self.channels.add_subject(subject_reflex_info["channels"])
            else:
                self.channels.add_subject()
            if "windows" in subject_reflex_info:
                self.windows.add_subject(subject_reflex_info["windows"])
            else:
                self.windows.add_subject()
            if "trials_sections" in subject_reflex_info:
                self.trials_sections.add_subject(subject_reflex_info["trials_sections"])
            else:
                self.trials_sections.add_subject()
            if "stim_params" in subject_reflex_info:
                self.stim_params.add_subject(subject_reflex_info["stim_params"])
            else:
                self.stim_params.add_subject()

        elif subject_reflex_info is None:
            self.channels.add_subject()
            self.windows.add_subject()
            self.trials_sections.add_subject()
            self.stim_params.add_subject()

    def init_section(self, section_id):
        """Initialise a section
        
        Parameters
        ----------
        section_id: str
            Specificy the name of the section to process.
            Section has to be one of the TextMark sections from Spike2 file.

        """
        self.section = section_id

        pkl = self.trial + "_" + self.section + ".pkl"
        self.section_pkl = self.study_path / self.subject / "proc" / pkl

        section_reflex_info = None
        try:
            json = self.trial + "_" + self.section + "_reflex.json"
            path = self.study_path / self.subject / "proc" / json
            section_reflex_info = s2pp.utils.read_json(path, strict=False)
        except FileNotFoundError:
            print(f'{json} does not exist')

        if section_reflex_info is not None:
            if "channels" in section_reflex_info:
                self.channels.add_section(section_reflex_info["channels"])
            else:
                self.channels.add_section()

            if "windows" in section_reflex_info:
                self.windows.add_section(section_reflex_info["windows"])
            else:
                self.windows.add_section()

            if "trials_sections" in section_reflex_info:
                self.trials_sections.add_section(section_reflex_info["trials_sections"])
            else:
                self.trials_sections.add_section()

            if "stim_params" in section_reflex_info:
                self.stim_params.add_section(section_reflex_info["stim_params"])
            else:
                self.trials_sections.add_section()

            if "rejected_trigger_windows" in section_reflex_info:
                self.rejected_trigger_windows = section_reflex_info[
                    "rejected_trigger_windows"
                ]

        elif section_reflex_info is None:
            self.channels.add_section()
            self.windows.add_section()
            self.trials_sections.add_section()
            self.stim_params.add_section()



import pytest
import spike2py_reflex as s2pr


@pytest.mark.plot
def test_outcome_summary_mmax_singles(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    s2pr.outcomes.summary(section)


# TODO: fix how fake data built to allow test to run; has to do with looking for section called 'mmax_2' but it is not specified in reflex windows
@pytest.mark.plot
def test_outcome_summary_mmax_single_two_muscles(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section.info.windows.ms_reflexes['Adm'] = section.info.windows.ms_reflexes[
        'Fdi']
    section.info.section = section.info.section + '_2'
    s2pr.outcomes.summary(section)


# TODO: fix how fake data built to allow test to run; has to do with looking for section called 'mmax_2' but it is not specified in reflex windows
@pytest.mark.plot
def test_outcome_summary_mmax_single_two_muscles_three_intensities(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    section.reflexes['Fdi'].avg_waveform[32] = 1
    section.reflexes['Fdi'].avg_waveform[35] = 1
    section.reflexes['Fdi'].mean_outcomes['mmax'][32] = section.reflexes['Fdi'].mean_outcomes['mmax'][1]
    section.reflexes['Fdi'].mean_outcomes['mmax'][35] = section.reflexes['Fdi'].mean_outcomes['mmax'][1]
    section.reflexes['Fdi'].mean_outcomes['hreflex'][32] = section.reflexes['Fdi'].mean_outcomes['hreflex'][1]
    section.reflexes['Fdi'].mean_outcomes['hreflex'][35] = section.reflexes['Fdi'].mean_outcomes['hreflex'][1]
    section.reflexes['Fdi'].reflexes[3].stim_intensity = 32
    section.reflexes['Fdi'].reflexes[2].stim_intensity = 32
    section.reflexes['Fdi'].reflexes[4].stim_intensity = 35
    section.reflexes['Fdi'].reflexes[5].stim_intensity = 35
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section.info.windows.ms_reflexes['Adm'] = section.info.windows.ms_reflexes[
        'Fdi']
    section.info.section = section.info.section + '_3'
    s2pr.outcomes.summary(section)


@pytest.mark.plot
def test_outcome_summary_hreflex_singles(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    s2pr.outcomes.summary(section)


# TODO: fix how fake data built to allow test to run; has to do with looking for section called 'mmax_2' but it is not specified in reflex windows
@pytest.mark.plot
def test_outcome_summary_hreflex_singles_two_muscles(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    section.reflexes['Adm'] = section.reflexes['Fdi']
    section.info.windows.ms_reflexes['Adm'] = section.info.windows.ms_reflexes['Fdi']
    section.info.section = section.info.section + '_2'
    s2pr.outcomes.summary(section)



@pytest.mark.plot
def test_outcome_summary_mmax_singles(info_data_ramp):
    info, data = info_data_ramp
    section = s2pr.reflexes.extract(info, data)
    section = s2pr.outcomes.calculate(section)
    s2pr.outcomes.summary(section)



import copy
from typing import Union
from dataclasses import dataclass

import numpy as np
from spike2py_preprocess.trial_sections import find_nearest_time_index

import spike2py_reflex as s2pr


@dataclass
class Window:
    single: Union[list, None] = None
    double: Union[list, None] = None
    double_single_pulse: Union[list, None] = None
    train_single_pulse: Union[list, None] = None


@dataclass
class WindowTypes:
    extract: Union[Window, None] = None
    plotting: Union[Window, None] = None
    reflexes: Union[dict, None] = None
    sd: Union[list, None] = None

    def clear(self):
        self.extract = Window()
        self.plotting = Window()
        self.reflexes = None
        self.sd = None


class GroupedWindows:
    """
    Windows and related info required to process data 

    Windows are defined in .json files, in ms. When the 
    sampling frequency and `double_isi` are provided, the
    various required indexes are computed.
    
    Attributes
    ----------
    fs: Union[float, int]
        data sampling rate
    x_axes: Window
        Computed x-axes for the various specified windows, 
        used for plotting and data extraction
    ms:
    idx
    double_isi
    idx_exists: bool
        Flag to track if idx have been computed


    """
    def __init__(self, win_info=None):
        """

        Parameters
        ----------
        win_info: dict 
            Info about windows provided in a .json file for the 
            study, subject or trial
        """
        self.fs = None
        self.x_axes = Window()
        self.ms = WindowTypes()
        self.ms.clear()  # To initialise
        self.idx = WindowTypes()
        self.idx.clear()  # To initialise
        self.double_isi = None

        self._idx_exists = False
        self._x_axis = None

        if win_info is not None:
            self.update(win_info)

    def clear(self):
        self.ms.clear()
        self.idx.clear()
        self._idx_exists = False
        self.x_axes = Window()

    def update(self, win_info):
        if "extract" in win_info:
            self.ms.extract = Window(**win_info["extract"])
        if "plotting" in win_info:
            self.ms.plotting = Window(**win_info["plotting"])
        if "reflexes" in win_info:
            self.ms.reflexes = win_info["reflexes"]
        if "sd" in win_info:
            self.ms.sd = win_info["sd"]

        if self.fs is not None:
            self._compute_idx_and_x_axes()

    def add_fs(self, fs):
        old_fs = self.fs
        self.fs = fs
        if old_fs != fs:
            self._get_x_axis()
            self._compute_idx_and_x_axes()

    def add_double_isi(self, double_isi):
        self.double_isi = double_isi

    def _compute_idx_and_x_axes(self):
        # TODO: see if this could be broken up into a few smaller methods
        # extract
        idx, x_axis = self._get_extract_idx_and_x_axis(self.ms.extract.single)
        self.idx.extract.single = idx
        self.x_axes.single = x_axis

        idx, x_axis = self._get_extract_idx_and_x_axis(self.ms.extract.double)
        self.idx.extract.double = idx
        self.x_axes.double = x_axis

        idx, x_axis = self._get_extract_idx_and_x_axis(self.ms.extract.double_single_pulse)
        self.idx.extract.double_single_pulse = idx
        self.x_axes.double_single_pulse = x_axis

        idx, x_axis = self._get_extract_idx_and_x_axis(self.ms.extract.train_single_pulse)
        self.idx.extract.train_single_pulse = idx
        self.x_axes.train_single_pulse = x_axis

        # reflexes
        self.idx.reflexes = dict()
        for emg_name, section_reflexes_windows in self.ms.reflexes.items():

            section_reflexes_with_window_idx = dict()

            for section, reflexes_windows in section_reflexes_windows.items():

                reflexes_with_window_idx = dict()
                for reflex_type_, reflex_ms in reflexes_windows.items():

                    idx1_single = find_nearest_time_index(self.x_axes.single, reflex_ms[0] * s2pr.utils.CONVERT_MS_TO_S)
                    idx2_single = find_nearest_time_index(self.x_axes.single, reflex_ms[1] * s2pr.utils.CONVERT_MS_TO_S)

                    idx1_double = find_nearest_time_index(self.x_axes.double, reflex_ms[0] * s2pr.utils.CONVERT_MS_TO_S)
                    idx2_double = find_nearest_time_index(self.x_axes.double, reflex_ms[1] * s2pr.utils.CONVERT_MS_TO_S)

                    try:
                        idx3_double = find_nearest_time_index(self.x_axes.double, (reflex_ms[0] + self.double_isi) * s2pr.utils.CONVERT_MS_TO_S)
                        idx4_double = find_nearest_time_index(self.x_axes.double, (reflex_ms[1] + self.double_isi) * s2pr.utils.CONVERT_MS_TO_S)
                    except TypeError:
                        print('Unable to compute double idx windows because double_isi not added to GroupWindows\n')
                        idx3_double = None
                        idx4_double = None

                    idx1_train = find_nearest_time_index(self.x_axes.train_single_pulse, reflex_ms[0] * s2pr.utils.CONVERT_MS_TO_S)
                    idx2_train = find_nearest_time_index(self.x_axes.train_single_pulse, reflex_ms[1] * s2pr.utils.CONVERT_MS_TO_S)

                    idx_for_reflex = {'single': [idx1_single, idx2_single],
                                      'double': [[idx1_double, idx2_double], [idx3_double, idx4_double]],
                                      'train': [idx1_train, idx2_train],
                                      }
                    reflexes_with_window_idx[reflex_type_] = idx_for_reflex
                section_reflexes_with_window_idx[section] = reflexes_with_window_idx
            self.idx.reflexes[emg_name] = section_reflexes_with_window_idx

        # sd
        self.idx.sd = dict()
        idx1_single = find_nearest_time_index(self.x_axes.single, self.ms.sd[0] * s2pr.utils.CONVERT_MS_TO_S)
        idx2_single = find_nearest_time_index(self.x_axes.single, self.ms.sd[1] * s2pr.utils.CONVERT_MS_TO_S)
        self.idx.sd['single'] = [idx1_single, idx2_single]
        self.idx.sd['double'] = None
        if self.double_isi is not None:
            idx1_double = find_nearest_time_index(self.x_axes.double, self.ms.sd[0] * s2pr.utils.CONVERT_MS_TO_S)
            idx2_double = find_nearest_time_index(self.x_axes.double, self.ms.sd[1] * s2pr.utils.CONVERT_MS_TO_S)
            self.idx.sd['double'] = [idx1_double, idx2_double]

        self._idx_exists = True

    def _get_extract_idx_and_x_axis(self, win: list):
        """Get indexes, relative to trig times (i.e. zero).

        Given the index of a trig time, apply the provided indexes to determine the indexes associated
        with the desired windows.

        Example: window for [-250ms , 50ms]
                 index of 0ms is 0
                 indexes for window are [-500, 100] # for a sampling of 2kHz
                 Thus, given the index of a trigger time (e.g. 1435) apply the window index
                 to determine data to extract (1435 + (-500), 1435 + 100)
         """
        zero_idx = find_nearest_time_index(self._x_axis, 0)
        lower_idx = find_nearest_time_index(self._x_axis, win[0] * s2pr.utils.CONVERT_MS_TO_S)
        upper_idx = find_nearest_time_index(self._x_axis, win[1] * s2pr.utils.CONVERT_MS_TO_S)
        idx = [lower_idx - zero_idx, upper_idx - zero_idx]
        x_axis = self._x_axis[lower_idx:upper_idx]
        return idx, x_axis

    def _get_x_axis(self):
        start = 1/self.fs
        stop = 1
        step = 1/self.fs
        positive_half_x_axis = np.arange(start, stop, step)
        negative_half_x_axis = -1 * (np.flip(positive_half_x_axis))
        middle = np.array([0])
        self._x_axis = np.concatenate([negative_half_x_axis, middle, positive_half_x_axis])


class Windows:
    """

    """

    def __init__(self, study_win_info):
        """

        Parameters
        ----------
        study_win_info
        fs
        """
        self._study = GroupedWindows(study_win_info)
        self._subject = GroupedWindows()
        self._subject_empty = True
        self._section = GroupedWindows()
        self._section_empty = True
        self.fs: Union[float, None] = None
        self.double_isi = None

    def add_subject(self, win_info=None):
        self._subject = copy.deepcopy(self._study)
        if win_info is not None:
            self._subject.update(win_info)
        self._subject_empty = False

    def add_section(self, win_info=None):
        self._section = copy.deepcopy(self._subject)
        if win_info is not None:
            self._section.update(win_info)
        self._section_empty = False

    def clear_subject(self):
        self._subject.clear()

    def clear_section(self):
        self._section.clear()

    @property
    def fs(self):
        return self.fs

    @fs.setter
    def fs(self, fs):
        self._study.add_fs(fs)
        if not self._subject_empty:
            self._subject.add_fs(fs)
        if not self._section_empty:
            self._section.add_fs(fs)

    @property
    def idx_sd(self):
        return self._section.idx.sd

    @property
    def idx_reflexes(self):
        return self._section.idx.reflexes

    @property
    def double_isi(self):
        if self._section.double_isi is not None:
            return self._section.double_isi
        if self._subject.double_isi is not None:
            return self._subject.double_isi
        if self._study.double_isi is not None:
            return self._study.double_isi
        else:
            return None

    @double_isi.setter
    def double_isi(self, double_isi):
        self._study.add_double_isi(double_isi)
        if not self._subject_empty:
            self._subject.add_double_isi(double_isi)
        if not self._section_empty:
            self._section.add_double_isi(double_isi)

    @property
    def idx_extract(self):
        return self._section.idx.extract

    @property
    def idx_plotting(self):
        return self._section.idx.plotting

    @property
    def x_axes(self):
        return self._section.x_axes

    @property
    def ms_sd(self):
        return self._section.ms.sd

    @property
    def ms_reflexes(self):
        return self._section.ms.reflexes

    @property
    def ms_extract(self):
        return self._section.ms.extract

    @property
    def ms_plotting(self):
        return self._section.ms.plotting
